/******************************************************************************************/
/* Coypright 2013, Winfried@sunkiddance.de aka  =-WDI-=, 28.12.2013                       */
/*                                                                                        */
/* This program is free software: you can redistribute it and/or modify                   */
/*     it under the terms of the GNU General Public License version 3 as published by     */
/*     the Free Software Foundation.                                                      */
/*                                                                                        */
/*     This program is distributed in the hope that it will be useful,                    */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of                     */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                      */
/*     GNU General Public License for more details.                                       */
/*                                                                                        */
/*     You should have received a copy of the GNU General Public License                  */
/*     along with this program.  If not, see <http://www.gnu.org/licenses/>.              */
/******************************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
  
#define bool int
#define true 1
#define false 0

#define SECTORSIZE 2048 /* DVD-Video sector size is 2048 bytes */
typedef enum _endian {little_endian, big_endian} EndianType;
/* Define data types regarding the ifo file */
typedef unsigned int ifo_4bytes;
typedef unsigned short ifo_2bytes;
typedef unsigned char ifo_byte;
typedef struct ADTEntry 
{ 
  ifo_2bytes VobID;    /* 1...n */
  ifo_4bytes startSec; /* 0...n */
  ifo_4bytes endSec;   /* 0...n */
} ADTEntry;

/* Describe kind of VOB sector type  */
typedef enum _sec {start, end} SecType;

/*------------------------------------------------------------------------------*/
/* Check endianess of CPU                                                       */
/* Code snipet shamelessly stolen from                                          */
/* http://www.rapidtables.com/prog/endianes                                     */
/*------------------------------------------------------------------------------*/ 
EndianType CheckCPUEndian(void)
{
    unsigned short x;
    unsigned char c;
    EndianType CPUEndian;
 
    x = 0x0001;
    c = *(unsigned char *)(&x);
    if( c == 0x01 )
        CPUEndian = little_endian;
    else
        CPUEndian = big_endian;
 
    return CPUEndian;
} /* end of CheckCPUEndian() */

/*------------------------------------------------------------------------------*/
/* Swap the endianess of an int (4 bytes!)                                      */
/* unsigned int x: integer which endianess should be converted                  */
/* return value: the "new" integer in the other endianess                       */ 
/* Code snipet shamelessly stolen from                                          */
/* http://www.rapidtables.com/prog/endianes                                     */
/*------------------------------------------------------------------------------*/ 
unsigned int EndianSwap32(const unsigned int x)
{
    unsigned int y=0;
    y += (x & 0x000000FF)<<24;
    y += (x & 0xFF000000)>>24;
    y += (x & 0x0000FF00)<<8;
    y += (x & 0x00FF0000)>>8;
    return y;
} /* end of EndianSwap32() */

/*------------------------------------------------------------------------------*/
/* Swap the endianess of an int (2 bytes!)                                      */
/* unsigned short x: short which endianess should be converted                  */
/* return value: the "new" short in the other endianess                         */ 
/* Code snipet shamelessly stolen from                                          */
/* http://www.rapidtables.com/prog/endianes                                     */
/*------------------------------------------------------------------------------*/ 
unsigned short EndianSwap16(const unsigned short x)
{
    unsigned short y=0;
    y += (x & 0x00FF)<<8;
    y += (x & 0xFF00)>>8;
    return y;
} /* end of EndianSwap16() */

/*------------------------------------------------------------------------------*/
/* Check the endianess of a value and swap the bytes if necessary               */
/* ifo_2bytes val: value to be checked                                          */
/* return value: the value in the proper endianess                              */
/*------------------------------------------------------------------------------*/ 
ifo_2bytes adaptEndianess16(ifo_2bytes val)
{
  EndianType CPUType;

  /* Check endianess of microprocessor */
  CPUType = CheckCPUEndian();
  if(CPUType == little_endian) return EndianSwap16(val); else return val;

} /* end of adaptEndianess16() */

/*------------------------------------------------------------------------------*/
/* Check the endianess of a value and swap the bytes if necessary               */
/* ifo_4bytes val: value to be checked                                          */
/* return value: the value in the proper endianess                              */
/*------------------------------------------------------------------------------*/ 
ifo_4bytes adaptEndianess32(ifo_4bytes val)
{
  EndianType CPUType;

  /* Check endianess of microprocessor */
  CPUType = CheckCPUEndian();
  if(CPUType == little_endian) return EndianSwap32(val); else return val;

} /* end of adaptEndianess32() */

/*------------------------------------------------------------------------------*/
/* Print the usage message                                                      */
/*------------------------------------------------------------------------------*/
void usage(void)
{
      fprintf(stderr, "VobFindSec 0.1 - Find start and end sectors of VOB's\n");
      fprintf(stderr, "Usage: vobfindsec ifo-file [VOBID]\n");
}

/*------------------------------------------------------------------------------*/
/* Issue error message via stderr                                               */
/* const char* message: format string for the message to print to stderr        */
/* const char* arg1: first % argument for the format string                     */
/* const char* arg2: second % argument for the format string                    */
/* bool showUsage: show usage information if true                               */
/*------------------------------------------------------------------------------*/
void handleError(const char* message, const char* arg1, const char* arg2, bool showUsage)
{
  if(message == 0)
    {
      fprintf(stderr, "handleError: Please provide at least an error message text!\n");
      return;
    }
  if(arg1 == 0 && arg2 == 0)
    {
      fputs(message, stderr);
    }
  else if((arg1 != 0 && arg2 == 0) || (arg1 == 0 && arg2 != 0))
    {
      fprintf(stderr, message, arg1 == 0 ? arg2 : arg1);
    }
  else
    {
      fprintf(stderr, message, arg1, arg2);
    }
  if(showUsage) usage();
}

/*------------------------------------------------------------------------------*/
/* Read the next VOB ID                                                         */
/* FILE* inputHandle: a file handle for a valid DVD-Video ifo file              */
/* ifo_4bytes secADT: starting sector address of cell address table             */
/* ifo_byte cellID: the cell id for the wanted VOB ID number                    */
/* Return value: the VOB ID                                                     */
/*------------------------------------------------------------------------------*/ 
ifo_2bytes getnextVOBID(FILE* inputHandle, const ifo_4bytes startSecADT, const ifo_byte cellID)
{
  ifo_2bytes VobID;
  int result;

  result = 0; /* This is the *error* condition */
  
  /* Read position of VOB id */
  if(fseek(inputHandle, startSecADT*SECTORSIZE+8+(cellID-1)*12, SEEK_SET) != 0 )
    {
      char bytePos[20]; /* Byte pos in ifo file to jump to, max value is 2^32*2^11+2^8 */
      snprintf(bytePos, sizeof(bytePos), "0x%x", startSecADT*SECTORSIZE+8+(cellID-1)*12);
      handleError("error: can't jump to position '%s': %s\n", bytePos, strerror(errno), false);
      return result;
    }

  if(fread(&VobID, 2, 1, inputHandle) != 1) 
    {
      handleError("error: can't read enough data: %s\n", strerror(errno), "", false);
      return result;
    }

  VobID=adaptEndianess16(VobID);

#ifdef DEBUG
  fprintf(stderr, "DEBUG: getnextVOBID: VOBID:  0x%x [%u]\n", VobID, VobID);
#endif
  
  return VobID;
} /* end of getNextVOBID() */

/*------------------------------------------------------------------------------*/
/* Find the sector number of the given cellID                                   */
/* FILE* inputHandle: a file handle for a valid DVD-Video ifo file              */
/* ifo_4bytes secADT: starting sector address of cell address table             */
/* ifo_byte cellID: the cell id for the wanted sector number                    */
/* SecType secType: start or end sector?                                        */
/* Return value: the sector number                                              */
/*------------------------------------------------------------------------------*/ 
ifo_4bytes getVOBSec(FILE* inputHandle, const ifo_4bytes secADT, const ifo_byte cellID, const SecType secType)
{
  ifo_4bytes secNum, sec; 
  int result; /* result in case of error */

  result = 4294967295; /* This is the *error* condition, 2^32-1 */
  
  /* Start or end sector? */
  switch(secType)
    {
    case start: 
      secNum = secADT*SECTORSIZE+8+4+(cellID-1)*12;
      break;
    case end:      
      secNum = secADT*SECTORSIZE+8+8+(cellID-1)*12;
      break;
    default:
      handleError("error: SecType has to be either 'start' or 'end\n", "", "", false);
      return result;
    }

  /* Read position of VOB sector */
  if(fseek(inputHandle, secNum, SEEK_SET) != 0 )
    {
      char secNumChar[20]; /* Sector number as string */

      snprintf(secNumChar, sizeof(secNumChar), "0x%x", secNum);
      handleError("error: can't jump to position '%s': %s\n", secNumChar, strerror(errno), false);
      return result;
    }

  if(fread(&sec, 4, 1, inputHandle) != 1) 
    {
      handleError("error: can't read enough data: %s\n", strerror(errno), "", false);
      return result;
    }

  sec=adaptEndianess32(sec);

#ifdef DEBUG
  fprintf(stderr, "DEBUG: getVOBSec: %5s sector:  0x%x [%u]\n", (secType == start ? "start" : "end"), sec, sec);
#endif

  return sec;
} /* end of getVOBSec() */

/*------------------------------------------------------------------------------*/
/* Find the start sector of the Cell Adress Data table within the ifo file      */
/* FILE* inputHandle: a file handle for a valid DVD-Video ifo file              */
/* char* ifoFile: the name of the ifo file                                      */
/* Return value: the sector number                                              */
/*------------------------------------------------------------------------------*/ 
ifo_4bytes findStartSecADT(FILE* inputHandle,  const char* ifoFile)
{
  ifo_4bytes startSecADT;
  int result = 4294967295; /* This is the *error* condition, 2^32-1 */

  /* Find start sector of video title set cells address table */
  if(fseek(inputHandle, 0xE0, SEEK_SET) != 0 )
    {
      handleError("error: can't jump to position '0xE0' in file '%s': %s\n", ifoFile, strerror(errno), false);
      return result;
    }

  /* Read start sector */
  if(fread(&startSecADT, 4, 1, inputHandle) != 1) 
    {
      handleError("error: can't read enough data in file '%s': %s\n", ifoFile, strerror(errno), false);
	  return result;
    }

#ifdef DEBUG  
  fprintf(stderr, "DEBUG: findStartSecADT: Start sector of ADT: %u\n", adaptEndianess32(startSecADT));
#endif

  return adaptEndianess32(startSecADT);
  
} /* end of findStartSecADT() */

/*------------------------------------------------------------------------------*/
/* Get the number of VOB's                                                      */
/* FILE* inputHandle: a file handle for a valid DVD-Video ifo file              */
/* char* ifoFile: the name of the ifo file                                      */
/* ifo_4bytes startSecADT: the start sector of the ADT                          */
/* Return value: the number of VOB's                                            */
/*------------------------------------------------------------------------------*/ 
ifo_2bytes getNumVOB(FILE* inputHandle, const char* ifoFile, ifo_4bytes startSecADT)
{
  ifo_2bytes numVOB;
  int result = 0; /* error */

  /* Go to video title set cells address table */
  if(fseek(inputHandle, startSecADT*SECTORSIZE, SEEK_SET) != 0 )
    {
      char bytePos[20]; /* Byte pos in ifo file */
      snprintf(bytePos, sizeof(bytePos), "0x%x", startSecADT*SECTORSIZE);
      handleError("error: can't jump to position '0x%x', %s\n", bytePos, strerror(errno), false);
      return result;
    }

  /* Read number of VOB's */
  if(fread(&numVOB, 2, 1, inputHandle) != 1) 
    {
      handleError("error: can't read enough date in file '%s': %s\n", ifoFile, strerror(errno), false);
      return result;
    }
  else
    {

#ifdef DEBUG
      fprintf(stderr, "DEBUG: getNumVOB: Number of VOB's: %u\n", adaptEndianess16(numVOB));
#endif

      return adaptEndianess16(numVOB);
    }

} /* end of getNumVOB() */

/*------------------------------------------------------------------------------*/
/* Read the data from the Cell Address Data Table into the ADTEntry struct      */
/* FILE* inputHandle: a file handle for a valid DVD-Video ifo file              */
/* ADTEntry* VobSec: Array of ADTEntries to be filled                           */
/* ifo_4bytes startSecADT: the start sector of the ADT                          */
/* ifo_2bytes numCells: the number of cells in the ADT                          */ 
/* Return value: the filled ADTEntry array                                      */
/*------------------------------------------------------------------------------*/ 
ADTEntry* buildVOBSecArray(FILE* inputHandle, ADTEntry* VobSec, const ifo_4bytes startSecADT, const ifo_2bytes numCells)
{
  ifo_byte cellID;
  ifo_2bytes curVobID, VobID=1;

  ADTEntry* result = NULL; /* error */

  /* Iterate the address table */
  for( cellID=1, curVobID=1, VobID=1; cellID <= numCells; cellID++)
    {
      if((VobID=getnextVOBID(inputHandle, startSecADT, cellID)) == 0)
	{
	  handleError("error: Can't read next VOB ID within cells address table\n", "", "", false);
	  return result;
	}
      if(VobID > curVobID)
	{
	  curVobID=VobID;
	  VobSec[VobID-1].VobID=VobID;
	  if((VobSec[VobID-1].startSec=getVOBSec(inputHandle, startSecADT, cellID, start)) == 4294967295)
	    {
	      handleError("error: Can't read start sector within cells address table\n", "", "", false);
	      return result;
	    }
	  if((VobSec[VobID-1].endSec=getVOBSec(inputHandle, startSecADT, cellID, end)) == 4294967295)
	    {
	      handleError("error: Can't read end sector within cells address table\n", "", "", false);
	      return result;
	    }
	}
      else 
	{
	  if(VobID == 1 && cellID == 1)
	    {
	      VobSec[VobID-1].VobID=VobID;
	      if((VobSec[VobID-1].startSec=getVOBSec(inputHandle, startSecADT, cellID, start)) == 4294967295)
		{
		  handleError("error: Can't read start sector within cells address table\n", "", "", false);
		  return result;
		}
	    }
	  if((VobSec[VobID-1].endSec=getVOBSec(inputHandle, startSecADT, cellID, end)) == 4294967295)
	    {
	      handleError("error: Can't read end sector within cells address table\n", "", "", false);
	      return result;
	    }
	}
    } /* end for */

  return VobSec;

} /* end of buildVOBSecArray() */

/*------------------------------------------------------------------------------*/
/* Find the end position sector of the Cell Address Data Table                  */
/* FILE* inputHandle: a file handle for a valid DVD-Video ifo file              */
/* char* ifoFile: the name of the ifo file                                      */
/* ifo_4bytes startSecADT: the start sector of the ADT                          */
/* Return value: the end sector number                                          */
/*------------------------------------------------------------------------------*/ 
ifo_4bytes findEndPosADT(FILE* inputHandle, const char* ifoFile, ifo_4bytes startSecADT)
{
  ifo_4bytes endPosADT;
  ifo_4bytes result=4294967295; /* error */

  /* Read end position of cell adress table */
  if(fseek(inputHandle, startSecADT*SECTORSIZE+4, SEEK_SET) != 0 )
    {
      char bytePos[20]; /* Byte pos in ifo file */
      snprintf(bytePos, sizeof(bytePos), "0x%x", startSecADT*SECTORSIZE);
      handleError("error: can't jump to position '0x%x': %s\n", bytePos, strerror(errno), false);
      return result;
    }

  if(fread(&endPosADT, 4, 1, inputHandle) != 1) 
    {
      handleError("error: can't read enough date in file '%s': %s\n", ifoFile, strerror(errno), false);
      return result;
    }

    return adaptEndianess32(endPosADT);

} /* end of findEndPosADT() */

/*------------------------------------------------------------------------------*/
/* Write the found sector start and end numbers to stdout                       */
/* ADTEntry* VobSec: the result                                                 */
/* ifo_2bytes outVOBId: the VOBId we are looking for, 0 if all are searched for */
/* ifo_2byte numVOB: the number of VOB's                                        */
/* Return value: none                                                           */
/*------------------------------------------------------------------------------*/ 
void outputResult(const ADTEntry* VobSec, const ifo_2bytes outVOBId, const ifo_2bytes numVOB)
{
  unsigned short i;

  if(outVOBId == 0) 
    {
      printf("VOBID\tStart\tEnd\n");
      for(i=0; i < numVOB; i++) 
	{
	  printf("%u\t%u\t%u\n", VobSec[i].VobID, VobSec[i].startSec, VobSec[i].endSec);
	}
    }
  else
    {
        i = outVOBId-1;
	printf("%u\t%u\t%u\n", VobSec[i].VobID, VobSec[i].startSec, VobSec[i].endSec);
    }
} /* end of outputResult() */
      
/*------------------------------------------------------------------------------*/
/* Find the sector number of the given VOB ID                                   */
/* ifoFile: a file name for a valid DVD-Video ifo file                          */
/* outVOBId: Id of the VOB we are looking for                                   */
/* Return value: 0 if successfull, 1 otherwise                                  */
/*------------------------------------------------------------------------------*/ 
int findSec(const char* ifoFile, ifo_2bytes outVOBId)
{
  FILE *inputHandle;
  int result;
  ifo_2bytes numVOB;           /* number of VOBs in ifo file */
  ifo_4bytes startSecADT;      /* starting sector of cell adress table */
  ifo_4bytes endPosADT;        /* end position of cells address table (byte number) */
  ifo_4bytes numCells;         /* number of cells in address table */
  ADTEntry* VobSec;            /* Dynamically allocated array of address table entries */
  ADTEntry* VobSecRes;         /* Dynamically allocated array of address table entries */

  result = 1;

  /* Open ifo file */
  if((inputHandle = fopen(ifoFile, "rb")) == 0) 
    {
      handleError("error: can't open input file '%s': %s\n", ifoFile, strerror(errno), false);
      return result;
    }

  /* Read start sector of ADT */
  if((startSecADT = findStartSecADT(inputHandle, ifoFile)) == 4294967295)
    {
      handleError("error: Can't find start sector of cell address table, file '%s': %s.\n", 
		  ifoFile, strerror(errno), false);
      fclose(inputHandle);
      return result;
    }

  /* Read number of VOB's */
  if((numVOB = getNumVOB(inputHandle, ifoFile, startSecADT)) == 0)
    {
      handleError("error: Can't find start sector of cell address table, file '%s': %s.\n", 
		  ifoFile, strerror(errno), false);
      fclose(inputHandle);
      return result;
    }

  /* Check 'VOBId' pararmeter */
  if(outVOBId > numVOB)
    {
      char outVOBIdChar[20]; 
      char numVOBChar[20];
      snprintf(outVOBIdChar, sizeof(outVOBIdChar), "%hu", outVOBId);
      snprintf(numVOBChar, sizeof(numVOBChar), "%hu", numVOB);
      handleError("error: VOBID '%s' too large, must be <= %s.\n", outVOBIdChar, numVOBChar, false);
      fclose(inputHandle);
      return result;
    }      
  /* Determine end byte number of ADT */
  if((endPosADT = findEndPosADT(inputHandle, ifoFile, startSecADT )) == 4294967295)
    {
      handleError("error: Can't find end byte number  of cell address table, file '%s: %s'.\n", 
		  ifoFile, strerror(errno), false);
	  fclose(inputHandle);
	  return result;
    }      

  /* Compute number of cells in address table */
  numCells=(endPosADT-7)/12;

#ifdef DEBUG
  fprintf(stderr, "DEBUG: Number of cells in address table: 0x%x [%u]\n", numCells, numCells);
#endif

  /* Allocate array of ADTEntry */ 
  if((VobSec=malloc(numVOB * sizeof *VobSec)) == NULL)
    {
      handleError("error: can't' allocate memory: %s\n", strerror(errno), "", false);
      fclose(inputHandle);
      return result;      
     }
  /* Build up a table of VOB id's  start - end sectors */
  if((VobSecRes = buildVOBSecArray(inputHandle, VobSec, startSecADT, numCells)) == 0)
    {
      handleError("error: can't build VOB sector table '%s': %s\n", ifoFile, strerror(errno), false);
      fclose(inputHandle);
      free(VobSec);
      return result;
    }

  /* Print the result */
  outputResult(VobSecRes, outVOBId, numVOB);

  /* Clean up */
  fclose(inputHandle);
  free(VobSec);

#ifdef DEBUG
  puts("DEBUG: Parsing .ifo-File, done\n");
#endif

  return 0;

} /* end of findSec() */

/*------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------*/
int main(int argc, char **argv)
{
  ifo_2bytes VOBId;         /* Which VOB Id to return? */
  char ifoFile[255]; /* the file name of the ifo file, the length chosen is arbitrary */

  VOBId=0; /* Default for returning all VOBs */

  if(argc != 2 && argc != 3) 
    {
      handleError("Error: Wrong number of arguments!\n", "", "", true);
      return 1;
    }
  else
    {
      if(sscanf(argv[1], "%254s", ifoFile) != 1)
	{
	  handleError("Error parsing 'ifo-file '%s' (string expected)\n", argv[1], "", true);
	  return 1;
	}
      if(argc == 3)
	{
	  if(sscanf(argv[2], "%10hu", &VOBId) != 1)
	    {
	      handleError("Error parsing 'VOBId '%s' (positive integer expected)\n", argv[2], "", true);
	      return 1;
	    }
	}
    }
 
  return findSec(ifoFile, VOBId);

} /* end of main() */ 
